import pdbfixer
import openmm.app as app
from pathlib import Path

def test_nonstandard():
    """Test adding hydrogens to nonstandard residues."""
    fixer = pdbfixer.PDBFixer(filename=(Path(__file__).parent / "data" / "4JSV.pdb").as_posix())
    fixer.removeChains(chainIndices=[0, 1, 2])
    fixer.addMissingHydrogens()
    for residue in fixer.topology.residues():
        count = sum(1 for atom in residue.atoms() if atom.element.symbol == 'H')
        if residue.name == 'ADP':
            assert count == 15
        if residue.name in ('MG', 'MGF'):
            assert count == 0

def test_leaving_atoms():
    """Test adding hydrogens to a nonstandard residue with leaving atoms."""
    fixer = pdbfixer.PDBFixer(filename=(Path(__file__).parent / "data" / "1BHL.pdb").as_posix())
    fixer.addMissingHydrogens()
    for residue in fixer.topology.residues():
        count = sum(1 for atom in residue.atoms() if atom.element.symbol == 'H')
        if residue.name == 'CAS':
            assert count == 10

def test_registered_template():
    """Test adding hydrogens based on a template registered by the user."""
    fixer = pdbfixer.PDBFixer(filename=(Path(__file__).parent / "data" / "1BHL.pdb").as_posix())

    # Register a template for CAS from which a single hydrogen has been removed.

    pdb = app.PDBFile((Path(__file__).parent / "data" / "CAS.pdb").as_posix())
    modeller = app.Modeller(pdb.topology, pdb.positions)
    modeller.delete([list(modeller.topology.atoms())[-1]])
    terminal = [atom.name in ('H2', 'OXT', 'HXT') for atom in modeller.topology.atoms()]
    fixer.registerTemplate(modeller.topology, modeller.positions, terminal)

    # See if the correct hydrogens get added.

    fixer.addMissingHydrogens()
    for residue in fixer.topology.residues():
        count = sum(1 for atom in residue.atoms() if atom.element.symbol == 'H')
        if residue.name == 'CAS':
            assert count == 9

def test_end_caps():
    """Test adding hydrogens to a chain capped with ACE and NME."""
    fixer = pdbfixer.PDBFixer(filename=(Path(__file__).parent / "data" / "alanine-dipeptide.pdb").as_posix())
    fixer.addMissingHydrogens()
    forcefield = app.ForceField('amber14/protein.ff14SB.xml')
    forcefield.createSystem(fixer.topology)


import pytest
from pdbfixer import PDBFixer
import openmm.unit


@pytest.mark.parametrize(
    "pdbCode,soluteCharge",
    [
        ("1PO0", -21),
        ("1A11", 1),
        ("110D", -5),
        ("1CNR", 0),
        ("1ESD", -21),
        ("25c8", -2),
    ],
)
def test_charge_and_solvate(pdbCode, soluteCharge):
    """
    Test that addSolvent successfully neutralises systems

    Parameters
    ----------
    pdbCode : str
        The PDB ID to test
    soluteCharge : int
        The formal charge of the solute - should equal the number of chloride
        ions minus the number of sodium ions in the neutralised system. Note
        that this may include ions from the original PDB entry.
    """
    fixer = PDBFixer(pdbid=pdbCode)
    fixer.findMissingResidues()

    chainLengths = [len([*chain.residues()]) for chain in fixer.topology.chains()]
    for chainidx, residx in list(fixer.missingResidues):
        if residx == 0:
            fixer.missingResidues[chainidx, residx] = ["GLY"]
        elif residx == chainLengths[chainidx]:
            fixer.missingResidues[chainidx, residx] = ["GLY"]

    fixer.findMissingAtoms()
    fixer.addMissingAtoms()
    fixer.addMissingHydrogens(pH=7.4)

    fixer.addSolvent(
        padding=2.0 * openmm.unit.nanometer,
        ionicStrength=0.1 * openmm.unit.molar,
        boxShape="dodecahedron",
    )

    numCl = sum(1 for res in fixer.topology.residues() if res.name.lower() == "cl")
    numNa = sum(1 for res in fixer.topology.residues() if res.name.lower() == "na")
    assert soluteCharge == numCl - numNa



import openmm.app as app
import pdbfixer
from pytest import raises
from pathlib import Path

def test_mutate_1():
    fixer = pdbfixer.PDBFixer(pdbid='1VII')
    fixer.applyMutations(["ALA-57-GLY"], "A")
    fixer.findMissingResidues()     
    fixer.findMissingAtoms()
    fixer.addMissingAtoms()        
    fixer.addMissingHydrogens(7.0)

    new_residue57 = list(fixer.topology.residues())[16]
    assert new_residue57.name == "GLY", "Name of mutated residue did not change correctly!"
    assert len(list(new_residue57.atoms())) == 7, "Should have 7 atoms in GLY 56"
    
    atom_names = set([atom.name for atom in new_residue57.atoms()])
    desired_atom_names = set(["N", "CA", "C", "O", "H", "HA3", "HA2"])
    assert atom_names == desired_atom_names, "Atom Names did not match for GLY 56"    


def test_mutate_2():
    fixer = pdbfixer.PDBFixer(pdbid='1VII')
    fixer.applyMutations(["ALA-57-LEU", "SER-56-ALA"], "A")
    fixer.findMissingResidues()     
    fixer.findMissingAtoms()
    fixer.addMissingAtoms()        
    fixer.addMissingHydrogens(7.0)
    new_residue57 = list(fixer.topology.residues())[16]
    new_residue56 = list(fixer.topology.residues())[15]
    assert new_residue57.name == "LEU", "Name of mutated residue did not change correctly!"
    assert new_residue56.name == "ALA", "Name of mutated residue did not change correctly!"
    
    assert len(list(new_residue56.atoms())) == 10, "Should have 10 atoms in ALA 56"
    assert len(list(new_residue57.atoms())) == 19, "Should have 19 atoms in LEU 57"
    
    atom_names = set([atom.name for atom in new_residue56.atoms()])
    desired_atom_names = set(["N", "CA", "CB", "C", "O", "H", "HA", "HB1", "HB2", "HB3"])
    assert atom_names == desired_atom_names, "Atom Names did not match for ALA 56"

    atom_names = set([atom.name for atom in new_residue57.atoms()])
    desired_atom_names = set(["C", "N", "CA", "CB", "CG", "CD1", "CD2", "O", "H", "HA", "HB2", "HB3", "HD11", "HD12", "HD13", "HD21", "HD22", "HD23", "HG"])
    assert atom_names == desired_atom_names, "Atom Names did not match for LEU 57"

def test_mutate_3_fails():
    with raises(ValueError):
        fixer = pdbfixer.PDBFixer(pdbid='1VII')
        fixer.applyMutations(["ALA-57-GLY", "SER-57-ALA"], "A")

def test_mutate_4_fails():
    with raises(KeyError):
        fixer = pdbfixer.PDBFixer(pdbid='1VII')
        fixer.applyMutations(["ALA-1000-GLY", "SER-56-ALA"], "A")

def test_mutate_multiple_copies_of_chain_A():
    fixer = pdbfixer.PDBFixer(pdbid='1OL5')
    fixer.applyMutations(['TPO-287-THR','TPO-288-THR'], "A")

def test_mutate_to_nonstandard():
    """Test mutating to a nonstandard residue defined with registerTemplate()."""
    fixer = pdbfixer.PDBFixer(filename=(Path(__file__).parent / "data" / "1BHL.pdb").as_posix())
    pdb = app.PDBFile((Path(__file__).parent / "data" / "CAS.pdb").as_posix())
    terminal = [atom.name in ('H2', 'OXT', 'HXT') for atom in pdb.topology.atoms()]
    fixer.registerTemplate(pdb.topology, pdb.positions, terminal)
    fixer.applyMutations(["SER-57-CAS", "ILE-60-CAS", "ASP-207-CAS"], "A")
    fixer.findMissingResidues()
    fixer.findMissingAtoms()
    fixer.addMissingAtoms()
    fixer.addMissingHydrogens(7.0)
    residues = list(fixer.topology.residues())
    for i in (0, 3, 134):
        assert residues[i].name == "CAS"
        atoms = list(residues[i].atoms())
        assert sum(1 for a in atoms if a.name == 'H2') == (1 if i == 0 else 0)
        assert sum(1 for a in atoms if a.name == 'OXT') == (1 if i == 134 else 0)
        assert sum(1 for a in atoms if a.name == 'HXT') == (1 if i == 134 else 0)

def test_download_template():
    """Test mutating to a nonstandard residue defined in the PDB."""
    fixer = pdbfixer.PDBFixer(filename=(Path(__file__).parent / "data" / "1BHL.pdb").as_posix())
    fixer.applyMutations(["SER-57-SEP", "ILE-60-SEP", "ASP-207-SEP"], "A")
    fixer.findMissingResidues()
    fixer.findMissingAtoms()
    fixer.addMissingAtoms()
    fixer.addMissingHydrogens(7.0)
    residues = list(fixer.topology.residues())
    for i in (0, 3, 134):
        assert residues[i].name == "SEP"
        atoms = list(residues[i].atoms())
        assert sum(1 for a in atoms if a.element == app.element.phosphorus) == 1
        assert sum(1 for a in atoms if a.name == 'OXT') == (1 if i == 134 else 0)

    # Check a few bonds to make sure the mutated residue has the ones it's supposed to.

    bonds = list(residues[3].bonds())
    assert sum(1 for a1, a2 in bonds if {a1.name, a2.name} == {'N', 'CA'}) == 1
    assert sum(1 for a1, a2 in bonds if {a1.name, a2.name} == {'CB', 'OG'}) == 1
    assert sum(1 for a1, a2 in bonds if {a1.name, a2.name} == {'P', 'OG'}) == 1
    assert sum(1 for a1, a2 in bonds if {a1.name, a2.name} == {'P', 'O2P'}) == 1